/*
 *   To compile:
 *
 *   gcc -fPIC -shared -o tigress_spank.so tigress_spank.c
 *
 */
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <mntent.h>
#include <sys/mount.h>
#include <sys/stat.h>

#include <slurm/spank.h>

/*
 *  * All spank plugins must define this macro for the
 *   * Slurm plugin loader.
 *    */
SPANK_PLUGIN(tigress_spank, 1);

#define FS_OFF 0
#define FS_ON 1
#define PROJECTS 2
#define TIGRESS 4

static int tigress_needed = 1;
static int projects_needed = 1;

static int _tigress_opt_process (int val,
                                const char *optarg,
                                int remote);

struct spank_option projects_opt =
    { "projects", NULL,
      "The job will need to use /projects and therefore will require nodes that have it mounted and available. This is the current default option.", 0, PROJECTS|FS_ON,
      (spank_opt_cb_f) _tigress_opt_process
    };

struct spank_option projects_off_opt =
    { "no-projects", NULL,
      "The job will not use /projects and therefore it can run on nodes that do not have it mounted/available. /projects will *not* be available to the job.", 0, PROJECTS,
      (spank_opt_cb_f) _tigress_opt_process
    };

struct spank_option tigress_opt =
    { "tigress", NULL,
      "The job will need to use /tigress and therefore will require nodes that have it mounted and available. This is the current default option.", 0, TIGRESS|FS_ON,
      (spank_opt_cb_f) _tigress_opt_process
    };

struct spank_option tigress_off_opt =
    { "no-tigress", NULL,
      "The job will not use /tigresss and therefore it can run on nodes that do not have it mounted/available. /tigress will *not* be available to the job.", 0, TIGRESS,
      (spank_opt_cb_f) _tigress_opt_process
    };

/*
int _tigress_print(const char *pr) {
    slurm_info("%s: tigress_spank, tigress=%d", pr, tigress_needed);
    slurm_info("%s: tigress_spank, projects=%d", pr, projects_needed);
}
*/

int slurm_spank_init (spank_t sp, int ac, char **av)
{
    int sc = spank_context();
    if (sc == S_CTX_ALLOCATOR || sc == S_CTX_REMOTE || sc == S_CTX_LOCAL) {
      spank_option_register (sp, &projects_opt);
      spank_option_register (sp, &projects_off_opt);
      spank_option_register (sp, &tigress_opt);
      spank_option_register (sp, &tigress_off_opt);
    }
    return (0);
}

// Use stat to get namespace number, expects full path, e.g. "/proc/1/ns/mnt"
int _tigress_get_namespace(const char *p,ino_t *num) {
    struct stat sb;

    if (stat(p, &sb) == -1) {
	slurm_error("tigress_spank: Failed to check namespace number of %s", p);
	return 1;
    }
    *num = sb.st_ino;
    return 0;
}

// Returns true/1 if filesystem p is mounted, 0 if not, -1 if failed
int _tigress_check_mount(const char *p) {
   struct mntent *ent;
   FILE *f;
   int ret=0;

   f = setmntent("/proc/mounts", "r");
   if (f == NULL) {
	slurm_error("tigress_spank: Failed to open /proc/mounts");
	return -1;
   }
   while (NULL != (ent = getmntent(f)) && !ret) {
	if (!strcmp(ent->mnt_dir, p)) {
	    ret = 1;
	}
   }
   endmntent(f);
   return ret;
}

// Check if p is mounted and if so attempt to unmount it
int _tigress_unmount(const char *p) {
    int is_mounted = _tigress_check_mount(p);
    if (is_mounted == -1) {
	return -1;
    }
    if (is_mounted == 0) {
    	slurm_debug2("tigress_spank: %s is already not mounted.", p);
	return 0;
    }
    // Looks to be still mounted, attempt unmount
    slurm_debug2("tigress_spank: %s is mounted, attempting unmount.", p);
    int rc = umount2(p, MNT_DETACH);
    if (rc) {
	if (rc==EBUSY) {
	    slurm_error("tigress_spank: Failed to unmount %s, it was busy. Attempting to continue.", p);
	    return 0;
	} else if (rc==EINVAL) {
	    // This could be because it was unmounted by another process, double check
	    if (!_tigress_check_mount(p)) {
		return 0;
	    } else {
		slurm_error("tigress_spank: Failed to unmount %s, probably locked.", p);
		return 0;
	    }
	} else {
	    slurm_error("tigress_spank: Failed to unmount %s, unknown error %d", p, rc);
	    return 1;
	}
    }
}

int slurm_spank_task_init_privileged(spank_t sp, int ac, char **av)
{
    ino_t one_ns = 0;
    ino_t self_ns = 0;
    if (_tigress_get_namespace("/proc/1/ns/mnt", &one_ns)) {
	return 1;
    }
    if (_tigress_get_namespace("/proc/self/ns/mnt", &self_ns)) {
	return 1;
    }
    if (one_ns == self_ns) {
	slurm_error("tigress_spank: Failed to run, running in same mount namespace as init process. Ignoring.");
	return 0;
    }
    slurm_debug3("tigress_spank: Detected init mnt ns %ju and my mnt ns is %ju", one_ns, self_ns);
    if (!tigress_needed) {
	if (_tigress_unmount("/tigress")) {
	    return 1;
	}
    } else {
	slurm_debug2("tigress_spank: /tigress has been requested, no unmount attempted.");
    }
    if (!projects_needed) {
	if (_tigress_unmount("/projects")) {
	    return 1;
	}
    } else {
	slurm_debug2("tigress_spank: /projects has been requested, no unmount attempted.");
    }
    // _tigress_print("slurm_spank_task_init_privileged");
}

static int _tigress_opt_process (int val,
                                const char *optarg,
                                int remote) {
    if (val & PROJECTS) {
	    projects_needed = val & 1;
    }
    if (val & TIGRESS) {
	    tigress_needed = val & 1;
    }
    return (0);
}
